// ----------------------------------------------------------------------------
// * Raima Database Manager
// *
// * Copyright (c) 2012 Raima Inc.,  All rights reserved.
// *
// * Use of this software, whether in source code format, or in executable,
// * binary object code form, is governed by the Raima LICENSE which
// * is fully described in the LICENSE.TXT file, included within this
// * distribution of files.
// * ----------------------------------------------------------------------------

#include <iostream>

// rdm.h is required in all files using the RDM Database
#include "rdm.h"

// obj17_db.h was generated from the DDL file obj17_db.ddl.
#include "obj17_db.h"

// obj17_db_api.h was generated from the DDL file obj17_db.ddl. with 
// the -cpp option
#include "obj17_db_api.h"

static uint8_t quit = 0;

// Name space for the RDM C++ interface
using namespace RDM_CPP;

// EXAMPLE - obj17 for C++
// This example shows multi-user access to a database using separate processes.
// and a remote tfserver.  The connections will use TCP/IP to communicate with
// the tfserver to do their work.  After starting the tfserver two writer 
// processes are created.  The first begins inserting values into
// the table.  The second waits until the table contains more than 100 rows
// and then begins deleting the oldest.  A reader process is also created
// that prints out the total number of records, the max value, and the min
// value.  Each thread uses read-only transactions (snapshots).

// The DDL (Database Definition Language) is in obj17_db.ddl. The database
// schema is embedded into the application and the database will be created
// the first time the application is run.  Subsequent runs will use the now
// existing database.

// The following RDM C++ methods are illustrated:
//      Db->Open
//      Db->Initialize
//      Db->BeginUpdate
//      Db->End
//      Db->Abort
//      Db->BeginSnapshot
//      Db->New_recordWithFieldValues
//      Db->Get_recordsByKey
//      Cursor->AfterLast
//      Cursor->GetFieldValues
//      Cursor++
// They are fully documented in http://docs.raima.com/rdm/.

// Each of these methods can throw and an exception of type rdm_exception

// The actual database is stored in a directory named 'obj17_db' in this
// project directory.

class obj17
{
public:
    Db_obj17_db    db;

    void InsertData () throw (rdm_exception&);
    void DeleteData() throw (rdm_exception&);
    void DisplayData() throw (rdm_exception&);
    void Initialize (uint16_t init_db=0) throw (rdm_exception&);
    void Cleanup () throw (rdm_exception&);
};

void obj17::Initialize(
    uint16_t init_db) throw (rdm_exception&)
{
    // Open the database and initialize it
    try
    {
        db = db.Open("obj17_db");
        if(init_db)
        {
            db.Initialize();
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") opening/initializing obj17_db database" << endl;
        throw (e);
        return;
    }
    return;
}

void obj17::InsertData() throw (rdm_exception&)
{
    int32_t      ii;
    struct rec01 rec01_rec;
    recordType rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};
        
    try
    {
        // Loop until there is an error or we are told to quit
        for (ii=0; !quit; ii++)
        {
            db.BeginUpdate(rtypes);

            rec01_rec.field01 = ii;
            db.New_rec01_recordWithFieldValues(&rec01_rec);

            db.End();
        }
    }
    catch (rdm_exception& e)
    {
        db.Abort();
        cerr << "Error " << e.error() << " (" << e.what() << ") adding records" << endl;
        throw (e);
        return;
    }
    return;
}

void obj17::DeleteData() throw (rdm_exception&)
{
    int32_t    ii;
    uint64_t   rec_count;
    recordType rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};
        
    try
    {
        // Loop until there is an error or we are told to quit
        for (ii=0; !quit; ii++)
        {
            db.BeginSnapshot();
            rec_count = db.Get_rec01_records().GetCount();
            db.End();

            if(rec_count > 100)
            {
                db.BeginUpdate(rtypes);
                db.Get_rec01_recordsBy_field01().First().Delete();
                db.End();
            }
            else
            {
                /* Wait for a bit to allow the inserter thread to get more records
                 * in the database
                 */
                sleep(1);
            }
        }
    }
    catch (rdm_exception& e)
    {
        db.Abort();
        cerr << "Error " << e.error() << " (" << e.what() << ") deleting records" << endl;
        throw (e);
        return;
    }
    return;
}
    
void obj17::DisplayData() throw (rdm_exception&)
{
    Cursor_rec01 rec01_cursor;
    uint64_t     rec_count = 0;
    int32_t      field01_min;
    int32_t      field01_max;
    char         msg[256];

    cout << endl;
    cout << "Total Records   Min Val  Max Val" << endl;
    cout << "-------------   -------  -------" << endl;

    try
    {
        // Loop until there is an error or we are told to quit
        while(!quit)
        {
            // Wait one second
            sleep(1);

            db.BeginSnapshot();

            rec01_cursor = db.Get_rec01_recordsBy_field01();

            rec_count = rec01_cursor.GetCount();

            rec01_cursor.First();
            rec01_cursor.Get_field01(&field01_min);

            rec01_cursor.Last();
            rec01_cursor.Get_field01(&field01_max);

            db.End();

            sprintf(msg, "       %6lu    %6d   %6d", rec_count, field01_min, field01_max);
            cout << msg << endl;
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") reading records" << endl;
        if(e.error() != S_UNAVAIL)
        {
            db.End();
        }
        return;
    }   
}

void obj17::Cleanup() throw (rdm_exception&)
{
    db.Release();
    return;
}

static void Deleter(void)
{
    TFS_HANDLE hTFS;
    obj17      obj;

    cout << "Deleter process" << endl;

    /* Initialize the application to use the external TFS (RPC) that we 
     * started
     */
    d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

    try
    {
        obj.Initialize();
        obj.DeleteData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in Deleter" << endl;
        exit(EXIT_FAILURE); 
    }

    d_tfsterm(hTFS);
}

static void Inserter(void)
{
    TFS_HANDLE hTFS;
    obj17      obj;

    cout << "Inserter process" << endl;

    /* Initialize the application to use the external TFS (RPC) that we 
     * started
     */
    d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

    try
    {
        obj.Initialize();
        obj.InsertData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in Inserter" << endl;
        exit(EXIT_FAILURE);
    }

    d_tfsterm(hTFS);
}

static void Reader(void)
{
    TFS_HANDLE hTFS;
    obj17      obj;

    cout << "Reader process (using read-only transactions)" << endl;

    /* Initialize the application to use the external TFS (RPC) that we 
     * started
     */
    d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

    try
    {
        obj.Initialize();
        obj.DisplayData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in Reader" << endl;
        exit(EXIT_FAILURE);
    }

    d_tfsterm(hTFS);
}

/* Start an external TFS Process */
static pid_t start_tfs_process(void)
{
    pid_t pid = fork();
    if (pid < 0)
    {
        cerr << "Error forking (" << strerror(errno) << ")" << endl;
        exit(EXIT_FAILURE);
    }

    if (pid == 0)
    {
        int32_t     stat;
        const char *tfserver_main_argv[] = {
            "tfserver"
        };

        (void) psp_init();

        stat = tfserver_main(1, tfserver_main_argv);
        if (stat != S_OKAY)
        {
            cerr << "Error starting tfserver (" << stat << ")" << endl;
            psp_term();
            exit(EXIT_FAILURE);
        }
    }

    return pid;
}

/* Stop a process */
static void stop_process(
    pid_t pid)
{
    int ret;
    int stat;

    /* Kill the tfserver process */
    ret = kill(pid, SIGINT);
    if (ret == -1)
    {
        cerr << "Error killing tfserver (" << strerror(errno) << ")" << endl;
        exit(EXIT_FAILURE);
    }

    /* Wait for the tfserver process to terminate */
    do
    {
        waitpid(pid, &stat, 0);
    } while (!WIFEXITED(stat) && !WIFSIGNALED(stat));
}

int32_t main(
    int32_t            argc,
    const char *const *argv)
{
    pid_t                inserter_pi;
    pid_t                deleter_pi;
    pid_t                reader_pi;
    pid_t                tfs_pi;
    TFS_HANDLE           hTFS;
    obj17                obj;

    cout << "Starting tfserver (and waiting 5 seconds for it to initialize)" << endl;

    /* Start the tfserver process to process 'remote' connections */
    tfs_pi = start_tfs_process();
    if (tfs_pi > 0)
    {
        /* Sleep to give the tfserver time to initialize and start listening */
        sleep(5);

        /* Initialize the application to use the external TFS (RPC) that we 
         * started */
        ::d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

        try
        {
            obj.Initialize(1);
        }
        catch (rdm_exception& e)
        {
            cerr << "Error " << e.error() << " (" << e.what() << ")" << endl;
            return EXIT_FAILURE;
        }

       /* Close the database in the main thread */
        obj.Cleanup();

        cout << endl << "Starting Processes..." << endl;
        cout << "Press <Enter> when you are ready to terminate them . . ." << endl;

        inserter_pi = fork();
        if (inserter_pi == 0)
            Inserter();
        else if (inserter_pi > 0)
        {
            deleter_pi = fork();
            if (deleter_pi == 0)
                Deleter();
            else if (deleter_pi > 0)
            {
                reader_pi = fork();
                if (reader_pi == 0)
                    Reader();
                else if (reader_pi > 0)
                {
                    /* Wait for user to hit <Enter> */
                    getchar();
    
                    ::d_tfsterm(hTFS);

                    stop_process(reader_pi);
                    cout << "Reader process finished" << endl;
                }

                stop_process(deleter_pi);
                cout << "Deleter process finished" << endl;
            }

            stop_process(inserter_pi);
            cout << "Inserter process finished" << endl;
        }

        /* Kill the tfserver process */
        stop_process(tfs_pi);
    }

    return EXIT_SUCCESS;
}
